#ifndef MCLASS_H
#define MCLASS_H

#include "mString.h"
#include "mDefs.h"
#include "mList.h"
#include "mStack.h"
#include "mMethod.h"
#include "mVariable.h"
#include <iostream>
#include "mLiteral.h"
#include "mOptions.h"
#include "mCode.h"
#include "mMethodTable.h"


class mMethod;
class mParseTree;
class mClosure;


/***********************************************************************************/

// variable is represented with: 
  //  - location in frame
  //  - availability (free or occupied)
  //  - the type of the variable
struct FrameVariable{
  int location;
  bool free;
  mClass *type;
  // meaningful for permanent literals only
  mLiteral *literal;
};


/***********************************************************************************/


// frame represetation
class FrameRecord{
  
  // frame contains parameters, named locals and then temporaries 
  // this is the index of 1st temporary variable
  int indexOfFirstTemp;  

  // list of variables in this frame
  mList<FrameVariable> *variables;

  // list of permanent literals in this frame
  mList<mSymbol> *permanentLiterals;

  // returns true if this is a permanent literal
  bool isPermanentLiteral(mLiteral*);

 public:
  // constructs new frame from variables in enviroment, method os owner method
  FrameRecord(mEnvironment*, mMethod*);
  virtual ~FrameRecord();

  // returns number of frame entries needed to hold all local (named and temporary)
  // variables
  int calculateSize();

  int result;          // final result (of the lastmost evaluation)
                       // result is either local variable or offset 
                       // in constant pool

  bool constant;       // true if result is an offset into constant pool
  

  // the method we are currently processing. this is useful because sometimes we need additional
  // info on method we're creating code for
  mMethod* ownerMethod;

  // now we define some useful flags

  // this is true iff we're invoking a method via variable reference, i.e.
  // myvar.method;
  bool flag_call_via_var;

  // this holds variable where (reference) value should be put
  // (used to set destination var for construction node)
  //int dest_variable;


  // returns first free frame variable and marks it 'occupied'
  int getFreeVariable();
  

  // frees occupied frame variable at specified location
  void free(int);

  // mark frame variable occupied
  void occupy(int);


  // shifts 1st temporary variable location
  void shiftTemporary(int offset) { indexOfFirstTemp += offset; }


  // frees occupied frame variable at specified location and cleans up stack if necessary
  //int free(int, mCode&, mString&);


  // finds (permanent) literal in frame and returns its frame variable 
  FrameVariable* findLiteral(mLiteral*);

  // makes variable at a specified location a permanent literal
  void bindPermanentLiteral(mLiteral*, int);

  // returns true if variable is a temporary
  bool isTemp(int);

  // returns string representation of result (offset or local var)
  mString strResult(){
    mString s;
    if( constant ) s = result; else { s='l'; s+= result; }
    return s;
  }
  
};


/***********************************************************************************/

// linkage table for external classes

// link table is built of rows of the following format:
// | className1 length | num entries | className1 | offset1 | offset2 | .... |
// | className2 length | num entries | className2 | offset1 | offset2 | .... |
// .............

// link table never contains reference for its own class because its address
// is already known
class linkTableEntry{
      public:
             linkTableEntry(mClass* c){
               theClass = c;  nOffsets = 0; offsets = 0;
             }
             
             virtual ~linkTableEntry(){
               if(offsets) delete[] offsets;
             }

             // adds offset to the list of offsets
             void addOffset(int offset){
               if(nOffsets == 0){
                  // 1st entry
                  offsets = new int[1];
                  offsets[0] = offset;
                  nOffsets++;
               }else{
                  // append to end
                  int *newOffsets = new int[nOffsets + 1];
                  for(int i=0;i<nOffsets;i++) newOffsets[i] = offsets[i];
                  newOffsets[nOffsets] = offset;
                  nOffsets++;
                  delete[] offsets;
                  offsets = newOffsets;
               }
             }
             
             // class
             mClass* theClass;

             // code offsets
             int* offsets;
             int nOffsets;   // # of offsets
};


// candidate method when searching for methods
      class candidateMethod{
          public:
	       candidateMethod(mMethod* m, int err){ method = m; error = err; }
	       mMethod*  method;     // the method
	       int error;            // error is 0 for exact match, and growing with inaccuracy
      };




// table of all objects that an object represents. i.e. for Integer this table would hold
// Object and Integer

class mObjectTableEntry{
 public:

	mObjectTableEntry(mClass *c, int offset){
	  theClass = c;
	  address = 0;    // address is not known at this point
	  selfishOffset = offset;
	  

	}

	mClass* theClass;    // the class
	int address;         // address
	int selfishOffset;   // selfish offset from self
};

class mObjectTable{

 public:
  mObjectTable(){

    entries = new mList<mObjectTableEntry>();
    entries->setAutoDelete( true );

  }

  virtual ~mObjectTable(){

    delete entries;

  }

  void addObject(mClass* cl, int offs){

    entries->append( new mObjectTableEntry(cl, offs) );

  }

  mList<mObjectTableEntry>* entries;	        	

};



// class
class mClass : public mSymbol{
      public:

             mClass();

             mClass(mString*, short);

             virtual ~mClass();
             
             // adds new method
	     // params: method to be added, class that caused error or 0
             bool addMethod(mMethod*, mClass*&);

	     // adds new method doesn't check compatibility with superclasses
	     // params: method to be added
             bool addMethod(mMethod*);


	     // adds superclass
             void addSuperClass(mClass*);

             // adds new field
             void addVariable(mVariable*);

             // returns string representation (without methods)
             mString toString();

             // returns string representation with all methods
             mString toStringFull();

	     // returns true if this is the class of a primitive type
	     // primitive types are Integer,Double,...
	     // this is used in casting to ascertain which cast must
	     // be performed (primitive or complex)
	     bool isPrimitive();


	     // exactly as isPrimitive but without Object
	     bool isPrimitiveButNoObject();

             // returns size of class
             int getSize();

	     // returns size of class info (size, mtab_offset, initial_selfish, ctable_size, meta_size)
	     // this is 20 bytes total
             int getInfoSize() { return 20; }
             
	     // returns size of object of this class in bytes
	     // this is the size that is to be allocated when instantiating
	     // this class (it's part of class info)
	     // 8 bytes is for object descriptor

	     int getObjectSize(){ 
	       // classptre + meta + method table + all descriptors, 8 bytes each
	       return 4 + 4 + methodTable->methodTableSizeOnly() + (methodTable->objectTableSize() / 12)*8 +
		 options->additionalObjectSize;  // + additional size at runtime for fields (Class, Method, and such)
	     }


	     // returns size of meta data for this class
	     int getMetaSize(){ 
	       return 4 +       // pointer to classname
		      4 +       // # of superclasses
		      4 +       // pointer to array of addresses of direct superclasses
		      4 +       // # of declared methods
		      4;        // pointer to table of method data
	     }

             // returns parameterless method with specified name, 0 if it
             // doesn't exist
             mMethod* getMethod(mString*);
	     

	     // returns method that most closely matches arguments. 0 if no such method has been found
	     // takes: name, arguments node, bool which when true searches superclasses
	     mMethod* getMethod(mString*, mParseTree*, bool, bool&);


	     // returns parameterless (default) constructor (note: there is
	     // only one default constructor per class)
	     mMethod* getDefaultConstructor();


	     // returns true if method already defined in this class. 
	     // only searches this class (without superclasses)
	     bool methodDefinedHere(mString*, mList<mParameter>*);

	     // returns mathod with specified signature. this is used for retrieving an already declared
	     // method
	     mMethod* getDeclaredMethod(mString*, mList<mParameter>*);


	     // changes return types and parameters of methods in all superclasses to this class
	     void updateSelfSuperTypes();

             // evaluates all class methods
             void evaluate();
             
	     // returns hieararhical distance between this and another class. 0 if the types are
	     // equal, -1 if the types are disjunctive (foreign)
	     // i.e. (object, integer) = 1, (integer, integer) = 0, (integer, some_user_type ) = -1
	     int distance(mClass*);

	     // returns true if this class is a subclass of another one
	     bool isSubclass(mClass*);

	     // returns true if this class is a subtype of another one. this includes Self
	     bool isSubtype(mClass*);


	     // returns index of link entry in link table for specified class
	     int getLinkEntry(mClass*);

	     short access;
	     mList<mMethod> *methodList;
	     mList<mSymbol> *literalList;
	     mList<mClass>  *superClasses;

	     mOptions* options;

	     // closures that are nested within 1st level closures (method bodies)
	     // used during evaluation
	     mStack<mClosure>* nestedClosures;

	     // closures that are nested within 1st level closures (method bodies)
	     // used to store them for relocation after method has been evaluated
	     mList<mClosure>* closures;

	     

	     // relocation table for constants
	     int *relocTable;
	     
	     // length of relocTable
	     int relocEntries;
	     
	     // linkage table for external classes
	     mList<linkTableEntry>* linkTable;
	     
	     // virtual & method table
	     mMethodTable* methodTable;

	     // table of all objects for this one
	     mObjectTable* objectTable;
	     
	     // this is true if call to method is ambiguous (used in error reporting when calling
	     // getMethod(...) )
	     bool ambiguous;

	     // true for special method types (for mMethodClass)
	     bool isMethodType;

	     // true if class has been preparsed
	     bool parsed;

	     // index of this class in class(object) table
	     int cTableIndex;	     	     

 private:            

	     // returns true if method is compatible with this class. this means that no such method
	     // must exist in class or signature of method must be compatible with existing method.
	     // the latter means that this method must be submethod of existing one
	     mClass* isCompatibleWith(mMethod*);

	     // fills a list of all possible methods in this class first and then in superclasses)
	     // returns true on success, false on error (ambiguity)
	     bool getPossibleMethods( mString*, mParseTree*,mList<candidateMethod>*,
				      mMethod*&, bool );

	     // fills a list of all possible methods in this class first and then in superclasses)
	     // returns true on success, false on error (ambiguity).
	     // used for parameterless mehods
	     bool getPossibleMethods( mString*, mList<candidateMethod>*,
				      mMethod*&, bool );
	     
	     // evaluates parse tree for a closure
	     void evaluateNode(mParseTree*, FrameRecord*, /*mCode&,*/ mEnvironment*);
	     
	     
	     // adds relocation entry for specified offset
	     void addRelocation(int);
	     
	     // adds offset to specified class in link table
	     void addLinkEntry(mClass*, int);
	     
	     // adds debug info to a string
	     mString debug(mString& code,const char *msg){
	       code += "  ( "; code += msg; code += " )";
	       return code;
	     }

	     // dumps all closures of a method
	     void dumpClosures(mEnvironment*);

	     // makes link entries for object table for all superclasses of this class
	     void linkSuperClasses(mClass*, mMethodTable*, int&);
	     
	     
	     // offset of code generated up to this moment
	     int currentCodeOffset;
	     
	     // address of this class' data is used when referencing this class
	     // so we don't add it to link table
	     int thisClassAddress;


	     // true if we have to generate status within block
	     bool requireStatus;

	     // true if evaluating arguments (needed by literal access)
	     bool evaluatingArguments;

	     // > 0 if updating a method. integer is needed because methods may be updated recursively
	     int updatingMethod;

	     // receiver class for invocation
	     mClass* receiver;
};


#endif
