#include "mClassDumper.h"
#include "mDefs.h"
#include "mLiteral.h"
#include <stdio.h>

// class file has te following structure
// -----------------------------
// |  head                     |   <-- class header
// -----------------------------
// |  reloc table              |   <-- relocation table
// -----------------------------
// |  link table               |   <-- linkage table for external classes
// -----------------------------
// |  literal pool             |   <-- literals used throughout the class
// -----------------------------
// |  meta data                |   <-- meta information for this class and its methods
// |   classname length        |
// |   classname               |
// |   # direct superclasses   |
// |  superindices into ctable |
// |  methods info 
// -----------------------------
// | class_data                |   <-- class table consists of information,
// | info: size,mtab_offset    |   virtual table(s) and method table
// |       initial_selfish     |   <-- offset from start of classdata
// | object table              |   <-- object table in the format | address | selfishOffset |
// | vtable_1                  |   <-- 1st virtual table \
// | vtable_2                  |    ........              \ viewed together
// | vtable_n                  |   <-- nth virtual table  /
// | method_signature_1, offset|  <-- method table
// | method_signature_2, offset|
// | .............             |
// | method_signature_n, offset|
// -----------------------------  <-- class code
// |  method1 code             |
// |  methodn code             |
// -----------------------------

// offseti so zaporedni naslovi metod (kot si sledijo v .class-u). zato je 
// treba tudi naslove v tej tabeli relocirat


using namespace std;

mClassDumper::mClassDumper(mClass* c)
{
  theClass = c;
}

mClassDumper::~mClassDumper()
{
}



void mClassDumper::dump()
{

 mString fname( *theClass->name );
 fname += ".class";
 FILE *fp = fopen( fname, "wb" );

 if( !fp ){
    cout<<"i can't write to "<<fname<<endl;
    return;
 }

 // setup header and dump it
 classHeader head; 
 strcpy( head.signature, "z0" );
 head.totalSize = 0;

 // # of methods in this class only (without superclasses) 
 head.nMethods = theClass->methodTable->methodTableEntries();

 // # of entries in virtual table
 head.vtableEntries = theClass->methodTable->virtualTableEntries();

 // # of entries in method table
 head.mtableEntries = theClass->methodTable->methodTableEntries();
 
 

 // # of entries in relocation table
 head.relocLength = theClass->relocEntries;

 // # of external classes in linkage table
 head.nExternalClasses = theClass->linkTable->size();

 // length of literal pool in bytes
 head.literalLength = 0;

 // length of metadata in bytes
 head.metaLength = theClass->getMetaSize();

 // # of closures
 head.nClosures = 0;

 // index of main method
 head.mainIndex = theClass->methodTable->getIndexOfMain(); 


 ///////////////////////////////////////////////////////////
 // header
 ///////////////////////////////////////////////////////////
 
 fwrite( &head, sizeof(classHeader), 1, fp );


 ///////////////////////////////////////////////////////////
 // reloc table
 /////////////////////////////////////////////////////////// 

 // reloc table is taken from class
 fwrite( theClass->relocTable, sizeof(int), theClass->relocEntries, fp );


 ///////////////////////////////////////////////////////////
 // linkage table
 /////////////////////////////////////////////////////////// 
 mList<linkTableEntry>* linkTable = theClass->linkTable; 

 
 // write link offsets for every external class (every row)
 for(linkTableEntry* entry=linkTable->first(); entry != 0; entry = linkTable->next() ){
   
   // format is: | class_name_length | num_entries | class_name | offset1 | offset2 | ....

   // class name length
   int length = entry->theClass->name->length();
   fwrite( &length, sizeof(int), 1, fp );
   
   // number of entries for this class
   fwrite( &entry->nOffsets, sizeof(int), 1, fp );

   
   // class name
   fwrite( *entry->theClass->name, length, 1, fp );

   // actual offsets
   if(entry->nOffsets > 0)
     fwrite( entry->offsets, sizeof(int), entry->nOffsets, fp );
   
 }


 ///////////////////////////////////////////////////////////
 // literal list (constant pool)
 ///////////////////////////////////////////////////////////
 mList<mSymbol>* literalList = theClass->literalList;

 for(mSymbol *c =  literalList->first(); c != 0; c = literalList->next()){
     mLiteral *lit = (mLiteral*)c;
     
     // Integer literal
     if(*lit->theClass->name == "Integer"){
        fwrite( &lit->intValue, sizeof(int), 1, fp );
        head.literalLength += 4;
     }
        
     // Double literal
     else if(*lit->theClass->name == "Double"){
        fwrite( &lit->doubleValue, sizeof(double), 1, fp );
        head.literalLength += 8;
     }
     // String literal
     else if(*lit->theClass->name == "String"){
       //char const* str = 

       // check whether this is native method name literal
       // if it is, we add null char
       if(lit->name->left(1) == 'Z'){   // native method names start with 'Z'
	 // write without '0' at the end
	 fwrite( *lit->stringValue, lit->length() - 1, 1, fp );
	 char nil = 0;
	 fwrite( &nil, 1, 1, fp );
	 head.literalLength += lit->length();	 
       }else{
	 fwrite( *lit->stringValue, lit->length(), 1, fp );
	 head.literalLength += lit->length();
       }
     }
     // Boolean literal
     else if(*lit->theClass->name == "Boolean"){
        fwrite( &lit->intValue, sizeof(char), 1, fp );
        head.literalLength += 1;
     }
 }




 ///////////////////////////////////////////////////////////
 // meta data for this class
 ///////////////////////////////////////////////////////////

 // length of class name + classname

 short classNameLength = theClass->name->length();
 fwrite( &classNameLength, sizeof(short), 1, fp );
 fwrite(*theClass->name, classNameLength, 1, fp );
 

 // number of direct superclasses
 short nSuperclasses = theClass->superClasses->size();
 fwrite( &nSuperclasses, sizeof(short), 1, fp );

 // array of addresses to superclasses. this will be linked at runtime, here
 // we set the addresses to indices into classtable
 
 int* addresses = new int[nSuperclasses];

 int index=0;
 theClass->methodTable->setObjectTableIndex(index);  // setup for all superclasses

 for(int i=0; i<nSuperclasses;i++){   
   
   addresses[i] = theClass->superClasses->at(i)->cTableIndex;   

 }

 fwrite( addresses, nSuperclasses, sizeof(int), fp );

 delete[] addresses;


 // methods declared in this class
 // each method is stored as a separate entry with the following layout:
 // method flags (byte),  methodname length, methodname, vtable index, # arguments, array of 
 // addresses to classdata for arguments, address to classdata for return type (optional)
 // bits in flags byte are: xxxxxxxx-- 1 if function (has return type)
 //                         |||||||--- 1 if static
 //                         ||||||---- 1 if native
 //                         |||||----- 1 if constructor
 //                         ||||------ y 00 private, 01 protected, 10 restricted, 11 public
 //                         |||------- y
 //                         ||-------- 1 if abstract
 //                         |--------- 1 if final
 // other bits are reserved


 for(mMethod *m =  theClass->methodList->first(); m != 0; m = theClass->methodList->next()){


   // setup method flags 
   unsigned char flags = 0;

   // 1st bit = 1 if this is a function
   if(m->isFunction())
     flags |= 1;

   // 2nd bit = 1 if it's static
   if(m->isStatic())
     flags |= 2;


   // 3rd bit = 1 if it's native
   if(m->isNative())
     flags |= 4;


   // 4th bit = 1 if it's constructor
   if(m->isConstructor())
     flags |= 8;


   // access modifiers in bits 5,6

   // private - 00 (don't do anything, bits are already cleared)
   if(m->access == ACCESS_PROTECTED)
     flags |= 0x10;  // set bit 5
   else if(m->access == ACCESS_RESTRICTED)
     flags |= 0x20;  // set bit 6
   else if(m->access == ACCESS_PUBLIC)
     flags |= 0x30;  // set bits 5,6
     
   

   fwrite( &flags, 1, 1, fp );



   // length of methodname
   unsigned char length = m->name->length();   
   fwrite( &length, 1, 1, fp );

   // methodname
   fwrite(*m->name , length, 1, fp );

   
   // index in virtual table
   int vtableIndex = theClass->methodTable->getMethodIndex(m);
   fwrite(&vtableIndex ,1, sizeof(int), fp );

   // object descriptor for this method
   int descriptor = theClass->methodTable->getDescriptor( m->theClass, 8 );
   fwrite(&descriptor ,1, sizeof(int), fp );

   // # of formal parameters
   unsigned char nParams = m->numParameters();   
   fwrite( &nParams, 1, 1, fp );
      

   
   for(mParameter* p = m->parameterList->first(); p != 0; p = m->parameterList->next()){

     int linkIndex = theClass->getLinkEntry( p->theClass );

     fwrite( &linkIndex, 1, sizeof(int), fp );

   }


   // index into link table for return type for functions
   if(m->isFunction()){

     int linkIndex = theClass->getLinkEntry( m->returnType );

     fwrite( &linkIndex, 1, sizeof(int), fp );

   }


 }


 ///////////////////////////////////////////////////////////
 // class data is built of class information, virtual
 // table(s) and method table
 ///////////////////////////////////////////////////////////

 ///////////////////////////////////////////////////////////
 // class information
 ///////////////////////////////////////////////////////////

 // object size   
 int object_size = theClass->getObjectSize();
 fwrite( &object_size, sizeof(int), 1, fp );

 // offset of method table (from start of class data)
 // that is: mtab_offset = class info (20 bytes) + vtable size + class table size
 int mtable_offset =  theClass->getInfoSize() + theClass->methodTable->virtualTableEntries()*4 + 
                      theClass->methodTable->objectTableSize(); 


 fwrite( &mtable_offset, sizeof(int), 1, fp );


 // initial selfish offset from start of class data
 int initial_selfish = theClass->getInfoSize() + theClass->methodTable->objectTableSize() /*+ 
                       ( theClass->methodTable->virtualTableEntries() - 
		       theClass->methodTable->thisVirtualTableEntries())*4*/;
 fwrite( &initial_selfish, sizeof(int), 1, fp );

 // class table size
 int ctable_size = theClass->methodTable->objectTableSize() / 12;   // number of entries !! not size in bytes
 fwrite( &ctable_size, sizeof(int), 1, fp );

 // size of meta data for this class
 int meta_size = theClass->getMetaSize();   // in bytes
 fwrite( &meta_size, sizeof(int), 1, fp );

 
 ///////////////////////////////////////////////////////////
 // object table
 ///////////////////////////////////////////////////////////

 dumpObjectTable( theClass->methodTable, fp );


 ///////////////////////////////////////////////////////////
 // virtual table(s)
 ///////////////////////////////////////////////////////////
 

 dumpVirtualTable( theClass->methodTable, fp );



 ///////////////////////////////////////////////////////////
 // method table
 ///////////////////////////////////////////////////////////

 dumpMethodTable( theClass->methodTable, fp );

 

 ///////////////////////////////////////////////////////////
 // the bloody code
 ///////////////////////////////////////////////////////////
 int codeSize = 0;
 for(mMethod *meth=theClass->methodList->first(); meth != 0; meth=theClass->methodList->next()){
   
   codeSize += dumpClosure( meth->environment->children->at(0), fp );
   
   
 } 


 // total size is size of file (without header, reloc and link table)
 // this is how much we need to hold class in memory
 // 20 is for class info
 head.totalSize = head.literalLength + theClass->getInfoSize() + theClass->methodTable->size() +
                  theClass->getMetaSize() + codeSize;

 // write updated header
 fseek( fp, 0, SEEK_SET );
 fwrite( &head, sizeof(classHeader), 1, fp );
 

 fclose( fp );

}


void mClassDumper::dumpVirtualTable(mMethodTable *table, FILE *fp)
{

  // parent tables
  for(mMethodTable *tab = table->superTables->first(); tab != 0; tab = table->superTables->next()){    
    //cout<<" dumping super table"<<endl;    
    dumpVirtualTable( tab, fp );

  }


  // for this class
  //cout<<" dumping own table"<<endl;
  for( virtualTable *vtab = table->virtualTables->first(); vtab !=0; vtab = table->virtualTables->next()){

    // write all entries for this vtable    
    for( virtualTableEntry *entry = vtab->entries->first(); entry !=0;  entry = vtab->entries->next()){

      fwrite( &entry->methodTableindex, sizeof(int), 1, fp );
      //cout<<"vtable: writen index "<<entry->methodTableindex<<endl;

    }
    //cout<<"----------------------"<<endl;
  }


}


void mClassDumper::dumpMethodTable(mMethodTable *table, FILE *fp)
{

  // parent tables
  for(mMethodTable *tab = table->superTables->first(); tab != 0; tab = table->superTables->next())    
    dumpMethodTable( tab, fp );


  // for this class
  for( methodTableEntry *entry = table->methodTable->first(); entry !=0; entry = table->methodTable->next()){
    int self=0;    
    fwrite( &entry->address, sizeof(int), 1, fp );  // address
    fwrite( &self, sizeof(int), 1, fp );            // place for actual self (now set to 0)
    fwrite( &self, sizeof(int), 1, fp );            // place for parent environment (now set to 0)

  }
}


void mClassDumper::dumpObjectTable(mMethodTable *table, FILE *fp)
{

  // parent tables
  for(mMethodTable *tab = table->superTables->first(); tab != 0; tab = table->superTables->next())    
    dumpObjectTable( tab, fp );


  // for this class
  // only one vtable should exist 
  for( virtualTable *vtab = table->virtualTables->first(); vtab !=0; vtab = table->virtualTables->next()){

    int address = 0;    // not known at this point
    fwrite( &address, sizeof(int), 1, fp );

    //cout<<"---- adding selfishOffset="<<vtab->selfishOffset<<endl;
    //cout<<"---- adding mtableOffset="<<vtab->mtableOffset<<endl;
    fwrite( &vtab->selfishOffset, sizeof(int), 1, fp );
    fwrite( &vtab->mtableOffset, sizeof(int), 1, fp );

      
  }

  
}



int mClassDumper::dumpClosure(mEnvironment *env, FILE *fp)
{
  mCode *code = env->code;

  // dump this environment's closure...
  fwrite( code->binary, code->binaryLength, 1, fp );

  int codeSize = code->binaryLength;
  
  // ... and all children
  for(mEnvironment* child = env->children->first(); child != 0; child = env->children->next()){
    
    codeSize += dumpClosure( child, fp );
    
  }

  return codeSize;
}

