#ifndef __M_ENVIRONMENT_H
#define __M_ENVIRONMENT_H

#include "mString.h"
#include "mList.h"
#include "mSymbol.h"
#include "mParseTree.h"
#include "mVariable.h"
#include "mParameter.h"
#include "mLiteral.h"
#include "mCode.h"
#include <iostream>

class mParseTree;
class mEnvironment{
      public:
             // creates new environment of specified type
             // type, name, parent
             // ( type can be method env, block env, ... )
             mEnvironment(short, mString*, mEnvironment*, mClass*);

             virtual ~mEnvironment();

             // appends new symbol
             void append(mSymbol*);

	     // appends new symbol to the list parent symbols. this is only called for parent
	     // symbol that is used in this environment
             void appendFromParent(mSymbol*);

	     // marks symbol as 'used' in this environmen't parental symbols and up every parent
	     // until owner of this symbol is reached
	     void markSymbolUsed(mString*);
             
             // returns string representation of this environment
             mString toString();

             // returns string representation of base environment
             // ( does not descend to children )
             mString toStringThis(int);

             // returns true if root environment (has no parent)
             bool isRoot(){ return parent == 0; }

             // adds new child environment
             void addChild(mEnvironment*);

             // returns symbol with specified name and type, 0 if it doesn't exist
             // only searches this environment
	     // type is the type of symbol we wish to search for (METHOD, VARIABLE, CLASS,...) 
	     // with that name or method with that name
	     // if type is -1 (default) then what ever comes first is returned (either
	     // variable or parameterless method).
	     // if type is SYMBOL_METHOD then we search for parameterless methods.
	     // methods with parameters are complex and are dealt with in mClass
             mSymbol* getLocalSymbol(mString*, int=-1);

             // returns symbol with specified name, 0 if it doesn't exist
             // searches this environment and its parents
	     // type is used in the same manner as in getLocalSymbol
	     // !! works for variables and parameterless methods !!
             mSymbol* getSymbol(mString*,int=-1);

	     // returns symbol already defined in this environment's parent symbols
	     // returns 0 if symbol was not found among parent symbols
	     mSymbol* getParentSymbol(mString*,int=-1);

             // returns first class - our current class
             mClass* getFirstClass();
	     
	     // returns declared class with specified name, 0 if it doesn't exist
	     // this is used for getting an already declared class in root environment
	     mClass* getDeclaredClass(mString*);

	     // returns declared class environment which is a child in this environment
	     // this is used for getting an already declared class environment in root environment
	     mEnvironment* getClassEnvironment(mString*);

	     // shifts variable locations by specified offset
	     void shiftVariables(int);

	     // updates locations of parent variables for this environment. this is needed
	     // because locations of parent variables may change during code generation but
	     // locations have been assigned during parsing
	     void updateParentVariables();

	     // finds specified symbol in actual parent, not parent symbols in this environment
	     // works for variables and parameters
	     mSymbol* findInParent(mString*);

	     // returns true if symbol with specified name was declared in direct parent
	     bool declaredInParent(mString*);

	     // returns true if symbol with specified name was declared in this environment
	     bool declaredHere(mString*);

	     // returns true if symbol with specified name was pushed as direct reference, not address 
	     // this function is used to test whether a symbol must be dereferenced first or if it
	     // can be used immediately. this is needed for method blocks
	     bool pushedDirectly(mString*);


	     // returns location of self in this environment. -1 if it doesn't exist
	     int getSelfLocation();


             // returns next free stack entry
             //int getNextVariableEntry();

             // creates all classes for builtin types (int, double,...)
             // and puts them in this environment
             void createBuiltinClasses();

	     // creates some additional or updates existing methods for builtin classes that could
	     // not be added because certain types were not yet available 
             void updateBuiltinClasses();

	     


	     // returns literal with specified name
	     mLiteral* findLiteral(mString*);

	     // returns permanent (used more than once) literal with specified name
	     mLiteral* findPermanentLiteral(mString*);

	     // removes literal from list of symbols
	     void removeLiteral(mLiteral*);


	     // adds self variable to this environment. this is normally called only in nonstatic method
	     // environments. parameter is self location
	     void addSelf(int);

	     // sets whether this environment has return statement
	     // args: true if return statement present, its line number
	     void setReturn(bool, int);

	     // sets whether this environment has parent assignment statement
	     // args: true if assignment present, its line number
	     void setParentAssignment(mClass*, bool, int);

	     // returns true if this environment's closure or any of its children has return statement
	     bool containsReturn();

	     // returns true if this environment's closure  or any of its children contain assignment
	     // statement to non primitive parent variable
	     bool containsParentAssignment();

	     // returns child of specified closure. only searches direct descendants!
	     // arg: closure
	     mEnvironment* getChildOfClosure(mParseTree*);
	


             // list of symbols in this environment (variables and literals)
             mList<mSymbol>* symbols;

	     // list of symbols from parent environment that are used in this environment
             mList<mSymbol>* parentSymbols;

	     // list of literals in this environment which occur multiple times
	     // note: every literal in this list is already in symbols list.
	     // this is used in parse tree evaluation to load multiply referenced
	     // literals only once and then keep them in frame
             mList<mSymbol>* permanentLiterals;
             
             // parent environment
             mEnvironment *parent;
             
             // list of children (can be more than 1)
             mList<mEnvironment> *children;           

             // every environment is basically a closure into which we append
             // statements and expressions
             mParseTree *closure;	     
             
             // environment type (method, class,...)
             short type;

             // name represents the name of a method, class or block
             // this is for debugging purposes only
             mString *name;

             // the class of this environment
             mClass *ownerClass;
             
             // current free entry on stack which can be used for
             // local variables
             int currentVariableEntry;

	     // code for this environment's closure	     
	     mCode *code;


	     // true if this environmen's closure contains a return statement.
	     bool hasReturn;

	     // true if this environmen's closure contains assignment to parent variable.
	     bool hasParentAssignment;

	     // true if this environment's closure is standalone (not assigned)
	     bool standalone;

	     // true if symbol that was last sought for is local
	     bool symbolIsLocal;

	     // true if this is method closure
	     bool methodClosure;

	     // code for closures nested within this environment
	     //mList<mCode>* nestedClosures;

             
};

#endif

