#ifndef MPARSETREE_H
#define MPARSETREE_H

#include "mString.h"
#include "mList.h"
#include "mSymbol.h"
#include "mVariable.h"
#include "mParameter.h"
#include <iostream>

class mClass;
class mMethod;
class mEnvironment;

class mParseTree{
      public:
             // creates empty parse tree of specified type and return type
             mParseTree(short, mClass* = 0);             
             // type, symbol, returnType
             mParseTree(short, mSymbol*, mClass* = 0);
	     // type, symbol name (this is for convenience only, it will be removed
	     // soon
             mParseTree(short, mString*, mClass* = 0);

             virtual ~mParseTree();

             // appends new subtree
             void append(mParseTree*);

             // returns string representation of this tree
             mString toString();

             // returns last method node in a sequence of method calls
	     // this only makes sense when tree represents method calls , i.e. a tree
	     // representing meth1.meth2.meth3 would return METHODCALL of meth3
             mParseTree *getLastMethodNode();

	     
	     // returns true if this tree ends with return statement. this makes sense only if this
	     // is closure tree (where last child is return)
	     bool endsWithReturn();

             // evaluates tree
             void evaluate();

	     // adds parameterless method call to the beggining of this tree's closure
	     void addSuperConstructorCall(mMethod*);


	     // returns string representation of this node's returnType even if it's set to 0 (void)
	     mString getType();


	     // returns true if this parsetree is a closure and contains a return statement
	     bool closureReturns();

	     // returns true if this parsetree is a closure and contains an assignment to non primitive
	     // parent variable
	     bool closureAssignsToParent();

	     // returns true if this parsetree is a closure with exactly 1 boolean expression
	     // used in loops
	     // args: class of Boolean
	     bool isBooleanClosure(mClass*);
	    

	     // returns number of if conditions in if staements
	     int numIfConditions();
	     
	     // promotes two primitive types to common type by inserting cast node
	     // and sets returnType of this parsetree accordingly
	     // i.e. if returnTypes of arguments are Integer and Double, Integer will
	     // be casted to Double and returnType of this tree set to Double
	     // this method is used by binary operations such as +,-,* given as
	     // 3rd argument
	     // 4th argument is optional. if it is present it's return type of this tree.
	     // returns false if cast could not be performed
	     bool promotePrimitivesViaCast(mParseTree*&, mParseTree*&, int, mClass* = 0);


	     // promotes single primitive type to desired return type by inserting cast node
	     // and sets returnType of this parsetree accordingly
	     // this method is used by unary operations such as !,- given as
	     // 3rd argument
	     // 4th argument is the desired type 
	     // returns false if cast could not be performed
	     bool promoteSinglePrimitiveViaCast(mParseTree*&, int, mClass*);

	     // promotes returnType of parsetree to new type if necessary by inserting
	     // cast node
	     // i.e. if returnType of this parsetree is Integer and desired type is 
	     // Double a cast Integer->Double will be inserted and new parsetree will 
	     // be returned
	     // works for primitive types only
	     bool promotePrimitiveViaCast(mParseTree*&, mClass*);

	     // if necessary performs a type cast from returnType of this parsetree
	     // to new type and inserts cast node into parsetree
	     // returns false if cast could not be performed, true on success
	     bool performTypeCast(mParseTree*&, mClass*);



	     // performs type cast on all arguments of a NODE_ARGUMENTS. this is valid
	     // only for parse trees representing arguments
	     // 2nd argument is method where we get types of formal parameters
	     // this function simply uses performTypeCast for every argument
	     // returns true on success
	     bool performArgumentsCast(mParseTree*&, mMethod*);

	     // returns string representation of NODE_PARAMETERS
	     mString toStringArguments();

	     // returns method invocation child if one exists, if argument is not -1, the child 
	     // must be located at specified position
	     mParseTree* getInvocationChild(int = -1);


	     // removes first child of specified node type (does not descend into children!)
	     // returns true if node was found, false otherwise
	     bool removeNode(short);


	     // returns true if this node is a variable reference
	     bool isVariableReference();

	     // returns true if this node is simple method reference, i.e. direct method call
	     bool isSimpleMethodReference();

	     // returns true if this tree is a nested methodcall, i.e. meth1.meth2....
	     bool isMultiMethodCall();

	     // returns true if this tree is a standalone (possibly nested) methodcall inside
	     // closure
	     // this is used to ascertain if a compound method call is standalone to set
	     // status inside conditional closures such as that for If statement
	     bool isStandaloneClosureMethodCall(mParseTree*);
	     
	     
	     // removes last method node in a multicall sequence
	     void removeLastMethodNode();	     


	     // returns true if ARGUMENTS_NODE contains no casts
	     bool argumentsMatchExactly();


	     // returns true if both types are primitives
	     static bool bothPrimitive(mParseTree*,mParseTree*);

             // list of children nodes
             mList<mParseTree>* children;
             
             // parent node
             mParseTree *parent;

             // node type
             short type;

	     // environment of this parsetree is needed in special cases
	     mEnvironment* env;

	     // class of return type is used in expressions and nodes that return
	     // something
	     mClass *returnType; 

	     // return type used for NODE_ANONMETHOD, for return type of anonymous function
	     mClass *functionReturnType; 


	     // used for literals, variables and everything that has to be stored
             mSymbol *symbol;
                           
	     // fully qualified identifiers, ...
	     // this will be removed later
             mString *name;

	     // used for casting expressions. cast is executed from 'fromType' to
	     // returnType
	     mClass *fromType;

	     // used for generating code when we need actual receiver of method
	     mClass *receiver;

	     // true for NODE_METHODCALL if self is passed implicitly
	     //  ... + meth(...);   this is implicit
	     //  ... + v.meth(...); this is explicit
	     // default is false 
	     bool implicitReference;

};

#endif
