#include "mVirtualMachine.h"
#include "mBuiltinClasses.h"
#include "mLiteral.h"
#include "mList.h"
#include "mDefs.h"
#include "mLinkEntry.h"
#include <string.h>
#include <dlfcn.h>    // shared libs






void dumpVirtualTable(int* table, int nEntries)
{

  for(int i=0;i<nEntries;i++)
    cout<<"vtable entry "<<i<<" = "<<table[i]<<endl;
}


void dumpMethodTable(int* table, int nEntries)
{

  for(int i=0;i<nEntries;i++)
    cout<<"mtable entry "<<i<<" : addr="<<table[3*i]<<", cself="<<table[3*i+1]
	<<", env offs="<<table[3*i+2]<<endl;
}


void dumpObjectTable(int* table, int nEntries)
{

  for(int i=0;i<nEntries;i++)
    cout<<"otable entry "<<i<<" : addr="<<table[3*i]<<", vtab offs="<<table[3*i+1]
	<<", mtab offs="<<table[3*i+2]<<endl;
}


/*********************************************************************************************/

mVirtualMachine::mVirtualMachine(char** args)
{  


  // allocate stack
  stack = (int*)malloc(stackSize); //new int[stackSize];

  int heapSize = 8000000;  // 8 MB  

  // setup memory
  vmHeap = originalHeap = (char*)malloc(heapSize);

  

  vmParameters = new mVMParameters( args );


  // create environment
  environment = new mVMEnvironment( this );
  


  // create native handles
  nativeHandles = new void* [ MAX_NATIVE_HANDLES ];


  // load all libs for native functionality
  loadNativeLibs();
  

  // this is where we store info for all loaded classes
  loadedClasses = new mList<mLoadedClass>();
  loadedClasses->setAutoDelete( true );


  // load classes for primitives
  loadPrimitiveClasses();
  
}

/*********************************************************************************************/


mVirtualMachine::~mVirtualMachine()
{

  delete environment;
  
  delete loadedClasses;  

  free(stack);  
  free(originalHeap);
  
  // ...and the libs
  unloadNativeLibs();

  
  delete vmParameters;


  delete[] nativeHandles;
}

/*********************************************************************************************/


mLoadedClass* mVirtualMachine::loadLocalClass(mString *path)
{
 FILE *fp; 

 ////////////////////////////////////////////
 // incorporate classFinder here
 ////////////////////////////////////////////

 // add .class

 mString classFile( *path );
 classFile += ".class";

 if((fp=fopen(classFile,"r"))==0){
      cout<<"cannot open "<<classFile<<endl;
      return 0;
 }


 // read header
 classHeader head;
 fread(&head, sizeof(classHeader), 1, fp );

 if( strcmp(head.signature,"z0" ) ){
    cout<<"invalid class file "<<endl;
    fclose(fp);
    return 0;
 }

 // check if class already loaded
 mLoadedClass* lClass = getLoadedClass( path );

 if( lClass != 0 ){ 
   fclose(fp);   
   return lClass;  
 }



 ///////////////////////////////////////////////////////////
 // allocate & read relocation table
 ///////////////////////////////////////////////////////////

 int* relocTable = new int[ head.relocLength ];
 fread( relocTable, sizeof(int), head.relocLength, fp );


 ///////////////////////////////////////////////////////////
 // allocate & read link table
 ///////////////////////////////////////////////////////////

 mLinkEntry** linkEntries = new mLinkEntry*[ head.nExternalClasses ]; 

 for( int i = 0; i < head.nExternalClasses; i++ ){

   // get class name and # of entries for it
   
   // format is: | class_name_length | num_entries | class_name | offset1 | offset2 | ....

   // lenght of classname in bytes
   int classNameLength = 0;
   fread( &classNameLength, sizeof(int), 1, fp );

   //cout<<"class_name len="<<classNameLength<<endl;
   
   // # of entries
   int nLinkEntries = 0;
   fread( &nLinkEntries, sizeof(int), 1, fp );

   
   // classname
   char *class_name = new char[ classNameLength + 1 ];
   fread( class_name, classNameLength, 1, fp );
   class_name[classNameLength] = '\0';   // add null
   mString cName(class_name);


   //cout<<"external class="<<class_name<<endl;
   // remove classname
   delete[] class_name;

   
   // read link offsets
   int* linkOffsets = new int[ nLinkEntries ];
   fread( linkOffsets, sizeof(int), nLinkEntries, fp );
     
        
   //cout<<"offsets are:";
   //for(int j=0;j<nLinkEntries;j++) cout<<linkOffsets[j]<<"  ";
   //cout<<endl;

   linkEntries[i] = new mLinkEntry( cName, nLinkEntries, linkOffsets );
   
 }

 
 
 ///////////////////////////////////////////////////////////
 // now we start with actual class data
 ///////////////////////////////////////////////////////////
 
 // allocate space for this class
 char* classSpace = new char[ head.totalSize ];


 // add this class to list of loaded classes
 mLoadedClass* loaded = new mLoadedClass( path, (unsigned)classSpace, head );
 loadedClasses->append( loaded );
 

 // first read literal pool
 fread( classSpace, head.literalLength, 1, fp );


 // read meta data and setup as pointers to separately alloc'd space
 /*int metaSize =*/ readMetaInfo( (int*)(classSpace + head.literalLength), head, fp );

 
 // read the rest of classfile (totalSize ensures we'll read till the end of file)
 fread( classSpace + head.literalLength + head.metaLength, head.totalSize, 1, fp );
 


 //////////////////////////////////////////////////////////
 // link class with external classes
 //////////////////////////////////////////////////////////

 for( int i = 0; i < head.nExternalClasses; i++ ){

   mLinkEntry* entry = linkEntries[i];

   // check if this class is already loaded. if it is we skip explicit loading
   lClass = getLoadedClass( entry->name );

   if( lClass != 0 ){
     
     entry->classAddress = lClass->dataStartAddress;
     // yes, it's loaded - skip loading

   }else{

     ////////////////////////////////////////////////
     // class is not loaded. load it up!
     ////////////////////////////////////////////////
     
     if( !(lClass=loadLocalClass( entry->name )) ){       

       // class could by some reason not be loaded. cleanup and return
       fclose(fp);
       delete[] linkEntries;
       delete[] relocTable;
       delete[] classSpace;
       return 0;
     }
     entry->classAddress = lClass->dataStartAddress;
     
   }  // class not loaded


   //cout<<" class is loaded at address "<<hex<<lClass->dataStartAddress<<endl;
   //cout<<"size is "<<*(int*)lClass->dataStartAddress<<endl;
   

   // external class is at this point available. now we have to link all references to this
   // external class in our class

   //cout<<"linking with external class"<<endl;
   linkExternalClass(classSpace, lClass->dataStartAddress, entry->offsets, entry->nOffsets);
      
 } 
   

 // relocate class code
 relocateClass( classSpace, (int)classSpace, relocTable, head.relocLength );
 delete[] relocTable;   // not needed any longer
  

 // relocate method addresses for THIS class only

 // virtual table starts after literal pool and method table starts after virtual table
 // |  literal pool   |   <-- starting address
 // |  class info     |
 // |  virtual table  |
 // |  method table   |


 ////////////////////////////////////////////////////////////////////////////////////////////
 // naslov te tabele pomeni zacetek tabele (VKLJUCNO z Objectom)!!!!!
 // moramo relocirati samo metode v delu tabele za NAS razred

 

 int ctableEntries = *(int*)(classSpace+ head.literalLength + head.metaLength + 12);


 //cout<<"--- object table  ( "<<ctableEntries<<" entries )" <<endl;
 int *classTable = (int*)(classSpace + head.literalLength + head.metaLength + 20);   // --> class table

 int initialSelf = 8;  // descriptor of 1st object goes after class and meta ptr
 

 for(int t=0;t<ctableEntries;t++){

   //cout<<"class address="<<*classTable<<", selfish offset="<<*(classTable+1)
   //  <<", mtable offset="<<*(classTable+2)<<endl;

    
   // relocate method table: this class_data, classtable entry
   //cout<<"relocating methodTable="<<(int)classTable<<endl;

 
   relocateMethodTable( classSpace + head.literalLength + head.metaLength, classTable, 
			(int)classSpace, initialSelf );   
   
   initialSelf += 8;

   classTable += 3;

 }


 // once again classtable ->
 classTable = (int*)(classSpace + head.literalLength + head.metaLength + 20);   // --> class table

 // link addresses of direct superclasses in meta section of superclasses
 int* metaData = (int*)(classSpace + head.literalLength);   // -> meta data 
 
 // 3rd field of metadata is ptr to a list of superclass addresses
 int* superList = (int*)*(metaData + 2);

 // 2nd field of metadata is # of direct superclasses
 for(int i=0; i < *(metaData + 1); i++){

   // get address for this class from ctable at position i*3 (3 entries for each entry) at 
   // position 0   
   superList[i] = classTable[ superList[i] * 3 + 0];

   
 }
 

 // link classaddresses of method parameters in meta section of methods

 MetaMethodEntry** entries = (MetaMethodEntry**) *(metaData + 4);

 // 4th field of metadata is # of methods
 for(int i=0; i <(int) *(metaData + 3); i++){
   

   int* paramAddresses = entries[i]->argClassAddresses;

   // check if we have function (includes return type)
   int withReturn = 0;
   if( entries[i]->flags & 1 != 0) withReturn = 1;

   for(int j=0; j<(int)entries[i]->nParameters + withReturn; j++){

     // get address of class in link table at position paramAddresses[j]          
     paramAddresses[j] = linkEntries[ paramAddresses[j] ]->classAddress;

   }     
   
   
 }
 
 
 // link offsets not needed anymore
 delete[] linkEntries;

 fclose(fp);

 return loaded;
}

/*********************************************************************************************/


mLoadedClass* mVirtualMachine::getLoadedClass(mString* className)
{

  for(mLoadedClass* lc = loadedClasses->first(); lc != 0; lc = loadedClasses->next()){
    
    if( *lc->name == *className ) return lc;

  }


  return 0;
}


mLoadedClass* mVirtualMachine::getLoadedClass(int address)
{

  for(mLoadedClass* lc = loadedClasses->first(); lc != 0; lc = loadedClasses->next()){
    
    if( lc->dataStartAddress == address ) return lc;

  }


  return 0;
}



mLoadedClass* mVirtualMachine::findAndLoadClass(mString* name)
{
  mLoadedClass* theClass = getLoadedClass( name );

  if(theClass != 0) return theClass;

  // otherwise we need to load it
  theClass = loadLocalClass( name );

  return theClass;
}


/*********************************************************************************************/


int mVirtualMachine::readMetaInfo(int* buffer, classHeader& head, FILE*& fp)
{

  int totalRead = 0;

  // read classname
  short length = 0;
  fread( &length, sizeof(short), 1, fp );
  totalRead += 2;


  char *className = new char[length + 1];   // +1 for '\0'
  fread( className, length, 1, fp );
  className[length] = '\0';

  totalRead += length;

  // store ptr to classname
  buffer[0] = (int)className;

  // # of direct superclasses
  short nSuperclasses = 0;
  fread( &nSuperclasses, sizeof(short), 1, fp );
  totalRead += 2;

  // store # of superclasses
  buffer[1] = nSuperclasses;

  // store ptr to array of superclass addresses

  int* superAddresses = new int[nSuperclasses];
  

  fread( superAddresses, nSuperclasses, sizeof(int), fp );
  totalRead += nSuperclasses*4;

  buffer[2] = (int)superAddresses;


  /////////////////////////////////////////////////
  // meta data for methods
  /////////////////////////////////////////////////
  
  // create method data table 
  // method table entry has the following layout
  // | flags | ptr to methodname | # arguments | ptr to array of classdata of all arguments |  


  MetaMethodEntry** entries = new MetaMethodEntry*[ head.nMethods ];
  

  for(int i=0; i < head.nMethods; i++){

    // read method flags
    unsigned char flags = 0;
    fread( &flags, 1, sizeof(char), fp );
    totalRead += 1;

    // methodname length
    unsigned char methodlength = 0;
    fread( &methodlength, 1, sizeof(char), fp );
    totalRead += 1;

    // methodname
    char *methodName = new char[methodlength + 1];   // +1 for '\0'
    fread( methodName, methodlength, 1, fp );
    methodName[methodlength] = '\0';
    totalRead += methodlength;

    // index in virtual table
    int vtableIndex;
    fread( &vtableIndex, 1, sizeof(int), fp );
    totalRead += sizeof(int);


    // descriptor
    int descriptor;
    fread( &descriptor, 1, sizeof(int), fp );
    totalRead += sizeof(int);


    // # parameters
    unsigned char nParams = 0;
    fread( &nParams, 1, sizeof(char), fp );
    totalRead += 1;
        

    // store metadata
    MetaMethodEntry *entry = new MetaMethodEntry;

    
    // 1 if it has return type (is function)
    int returnTypeEntry = flags & 1 != 0 ? 1 : 0;

    entry->flags = flags;
    entry->methodName = methodName;
    entry->vtableIndex = vtableIndex;
    entry->descriptor = descriptor;
    entry->nParameters = nParams;
    // + 1 entry for functions because of return type
    entry->argClassAddresses = new int[nParams + returnTypeEntry];


    // read classdata for all parameters
    fread( entry->argClassAddresses, nParams + returnTypeEntry, sizeof(int), fp );
    totalRead += (nParams + returnTypeEntry) * sizeof(int);

    //cout<<"params for "<<methodName<<endl;
    //for(int j=0;j<nParams + returnTypeEntry; j++)
    //cout<<"Param index="<<entry->argClassAddresses[j]<<endl;


    entries[i] = entry;    
    
  }

  
  buffer[3] = head.nMethods;   // # of declared methods

  buffer[4] = (int)entries;   // -> methodinfo table
  
  return totalRead;
}


/*********************************************************************************************/

void mVirtualMachine::relocateClass(char* classData, int address,
     int* relocTable, int relocEntries)
{

 for(int i = 0; i < relocEntries; i++){
    *(int*)(classData + relocTable[i]) += address;
 }
}


/*********************************************************************************************/




void mVirtualMachine::relocateMethodTable(char* classData, int* classTableEntry, int thisAddress, 
					  int initialSelf)
{
  // we relocate portion of method table of class pointed to by classData with method table
  // from classTableEntry


  int* otherData = (int*)*classTableEntry;  // class data of other class
  

  // first we need header of other class
  mLoadedClass* loaded = getLoadedClass( (int)otherData );

  if(!loaded) return;  // this should never occur 


  

  int thisStartingOffset = *(classTableEntry+2);  // offset into method table of our class

  // now we have to copy method table from other class to method table of our class
  // from otherData

  // starting offset of method table of other class is stored in last classtable entry of other class

  int otherStartingOffset = *(otherData + 5 + ( *(otherData + 3) - 1) * 3 + 2);

  //cout<<"other object table entries "<<*(otherData + 3)<<endl;  
  
 
  //cout<<"thisStartingOffset="<<thisStartingOffset<<endl;
  //cout<<"otherStartingOffset="<<otherStartingOffset<<endl; 
  //cout<<"other class has "<<loaded->nMethods<<" methods"<<endl; 


  if( (int)classData == (int)otherData ){


    // if this is our own class we relocate addresses
    //cout<<"--- linking my own class"<<endl;
    //cout<<"mtable offset ="<<*(int*)(classData + 4)<<endl;

    //dumpObjectTable((int*)(classData + 16), 2);


    //dumpVirtualTable((int*)(classData + 16+24), 7);
    

    for(int i = 0; i < loaded->nMethods*3; i+=3){

      *(int*)(classData+ thisStartingOffset+ (i<<2)) += thisAddress;
      *(int*)(classData+ thisStartingOffset+ (i<<2) + 4) = initialSelf;

      
    }

    //dumpMethodTable((int*)(classData + 16+24 + 7*4), 7);


  }else{

    /*cout<<"--- linking object"<<endl;
    cout<<"copying from "<<(int)((char*)otherData+otherStartingOffset)<<" to "
	<<(int)((char*)classData+thisStartingOffset)<<", "<<loaded->nMethods*12<<" bytes"<<endl;
    */
    

    // otherwise we copy entire method table from superclass
    memcpy( classData + thisStartingOffset, (char*)otherData+otherStartingOffset, loaded->nMethods*12 );

    for(int i = 0; i < loaded->nMethods*3; i+=3){
      //cout<<"method "<<i<<" address is "<<*(int*)(classData + thisStartingOffset+ (i<<2))<<endl;

      *(int*)(classData+ thisStartingOffset+ (i<<2) + 4) = initialSelf;
      
    }

  }

  


}



/*********************************************************************************************/


void mVirtualMachine::linkExternalClass(char* classData, int address, int* offsets, int nOffsets)
{

  for(int i = 0; i < nOffsets; i++){
    
    *(int*)(classData + offsets[i]) = address;

  }

}

/*********************************************************************************************/


void mVirtualMachine::executeClass(mLoadedClass* cl)
{

  // check if class has method main
  if(cl->head.mainIndex == -1){
    cout<<"class "<<*cl->name<<" has no method main"<<endl;
    return;

  }

  // create interpreter
  
  interpreter = new mTestInterpreter( this );
 
  // execute method main of this class
  interpreter->execute( cl );

  

}




int* mVirtualMachine::create_object(int *classData)
{
  
       int size = *classData;       


       int* object = (int*)allocate( size ); 

       /* store classdata ptr*/ 
       *object = (int)classData;


       //*(object+1) = metadata 
   
       /* store Object descriptor */ 
       *(object+2) = (int)classData + object_vtable_offset; //*(classData+2); 
       *(object+3) = (int)object;  

       /* store Other (Integer, Double,...) descriptor */ 
       *(object+4) = (int)classData + second_vtable_offset; // *(classData+2); 
       *(object+5) = (int)object; 
        

       // 4*12 = 4 method entries in mtable












       //////////////////////////////////////////////////////////
       // 5 METOD IMAMO!!!!!!!!!!!!!!
       // SPREMENI TO!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       //////////////////////////////////////////////////////////



       // setup method table


       // copy from classdata into object
       memcpy( object + 6, classData + (*(classData+1) >> 2) , 4*12 /*size - 24*/ );
       
       // and set actual self and environment
       for(int i=1; i <= (/*size - 24*/4*12)>>2; i+=3){ 

	 //////////////////////////////////////////////////////
	 ///////////////////////////////////////////////////////
	 // tukaj moramo nastavit na zacetni objekt, ki je Integer, Double, String,... in
	 // ne Object!!!!
	 // metode bodo morale biti prevedene KOT v originalni tabeli (Object.print -> #1,
	 // String.print -> #1 in NE v #3 !!!!)

	 *(object + 6 + i) = (int)(object + 4);   // actual self  + 4 TUKAJ!!!!!!!!!!!!!1
	*(object + 6 + i+1) = 0;  
       } 
       

       return object;

}



int* mVirtualMachine::create_user_object(int *classData, int addSize)
{

  int size = *classData;

  // alloc space
  int* object = (int*)allocate( size + addSize );
     
  *object = (int)classData;     // --> points to classdata
  *(object+1) = 0;              // --> points to meta object



  // create descriptors
  int nDescriptors = *(classData+3);


  // each descriptor is a pair: selfish | self
  // selfish points to virtual table inside classdata and self points to start of object : int* object
  for(int i=0;i<nDescriptors;i++){
    // offset of vtable is taken from classtable in classdata which begins after classinfo
    // classtable format is: address | selfish offset | method table offset
    // 5 is for classinfo (which is 5 * int) 
    *(object + 2 + 2*i) = (int)classData + *(classData + 5 + 3*i + 1);   // to vtable in classdata
    *(object + 2 + 2*i+1) = (int)object;    // original object
	
  }


  // thisObject points to 'end' object of created object. If Test object which inherits Object was
  // created then thisObject would point to descriptor for Test and not Object. 
  // descriptor for the end object is always the last descriptor because the class of this object
  // is the last class in hierarchy
  int thisObject = (int)(object + 2 + (nDescriptors-1)*2);


  // copy method table into this object from classdata
  // destination is this object after class ptr, meta, and all descriptors
  // source is methodtable, the offset of which is given in 2nd field of classinfo (*(classData+1)),
  // >> 2 for bytes to # ints 
  // method table size is: total object size - 4 (classdata ptr) - 4 (meta ptr) - # descriptors * 8 bytes each
  memcpy( object + 2 + nDescriptors*2, classData + (*(classData+1) >> 2) , size - 8-nDescriptors*8 );  


  // TUKAJ je treba nastavit deskriptorje za nadrazrede. NADRAZREDI NIMAJO istega deskriptorja kot
  // zdaj - thisObject


  // setup each entry in method table
  // the number of 12 byte entries is size / 4
  for(int i=1; i <= (size - 8-nDescriptors*8)>>2; i+=3){
    

    
    //TUKAJ
    //ZDAJ DELA TEST.Z0 (DEDOVANJE), DRUGACE PA NE DELA (CE JE SAMO = thisObject)
    //cout<<"default self for method="<< *(object + 2 + nDescriptors*2 + i)<<endl;

    // actual self = object
    
    *(object + 2 + nDescriptors*2 + i) += (int)object; //thisObject;  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // environment ptr is set to 0 
    *(object + 2 + nDescriptors*2 + i+1) = 0;  
  }

    
  return (int*)thisObject;
  
}




/*********************************************************************************************/


void* mVirtualMachine::allocate(int size  /*, classNumber for allocation*/)
{
  // space will be allocated from object pool:

  // | Integer | Double | String | MyClass A | MyClassB | ....
  //    |                   |          |---------------------> .....          
  //    |                   |--> space for String objects
  //    --> space for Integer objects


  // but for now just allocate desired space and return it
  void* space = (void*)vmHeap; //malloc( size );
  vmHeap += size;
  return space;
}



void mVirtualMachine::loadPrimitiveClasses()
{
  

  // virtual primitive methods

  // the following are implementations of methods for some builtin classes, such as Integer, Double, 
  // Block, etc. all methods that may be overriden in subclasses must be implemented to be
  // virtually invoked. all other methods have their implementation in VM itself (via instructions)

  // Object.print
  // print: new_frame_copy 3,3; Object.print l3; return_1
  char *object_print = new char[18];
  int len =  appendInstruction(object_print, 0, ClassCore::NEW_FRAME_COPY, 3, 3);
  len += appendInstruction(object_print, len, ClassObject::PRINT, 3);
  len += appendInstruction(object_print, len, ClassCore::RETURN_1);  


  // Integer.print
  // print: new_frame_copy 3,3; Integer.print l3; return_1
  char *integer_print = new char[18];
  len =  appendInstruction(integer_print, 0, ClassCore::NEW_FRAME_COPY, 3, 3);
  len += appendInstruction(integer_print, len, ClassInteger::PRINT, 3);
  len += appendInstruction(integer_print, len, ClassCore::RETURN_1);

  
  // Double.print
  // print: new_frame_copy 3,3; Double.print l3; return_1
  char *double_print = new char[18];
  len =  appendInstruction(double_print, 0, ClassCore::NEW_FRAME_COPY, 3, 3);
  len += appendInstruction(double_print, len, ClassDouble::PRINT, 3);
  len += appendInstruction(double_print, len, ClassCore::RETURN_1);


  // String.print
  // print: new_frame_copy 3,3; String.print l3; return_1
  char *string_print = new char[18];
  len =  appendInstruction(string_print, 0, ClassCore::NEW_FRAME_COPY, 3, 3);
  len += appendInstruction(string_print, len, ClassString::PRINT, 3);
  len += appendInstruction(string_print, len, ClassCore::RETURN_1);

  // Boolean.print
  // print: new_frame_copy 3,3; Boolean.print l3; return_1
  char *boolean_print = new char[18];
  len =  appendInstruction(boolean_print, 0, ClassCore::NEW_FRAME_COPY, 3, 3);
  len += appendInstruction(boolean_print, len, ClassBoolean::PRINT, 3);
  len += appendInstruction(boolean_print, len, ClassCore::RETURN_1);



  // Block.isTopLevel is implemented in native, so we just generate code for
  // native invocation here
  // print: new_frame_copy 4,4; Core.invoke_n_f literal_add, 4
  char *block_isTopLevel_literal = "Z__Block__isTopLevel__0";   // method name in native (Block.cc)
  char *block_isTopLevel = new char[20];
  len =  appendInstruction(block_isTopLevel, 0, ClassCore::NEW_FRAME_COPY, 4, 4);
  len += appendInstruction(block_isTopLevel, len, ClassCore::INVOKE_N_F, (int)block_isTopLevel_literal, 4);



  ///////////////////////////////////////////////////////////////////////////////////////////
  // TE metode se preveri!!!
  // metoda je dodana samo, kadar bo klicana preko invokacije. Npr. exec pri Bloku, While,...
  // NE dodamo, ker nima svojega indeksa v tabeli
  // primitivne metode se pri prevajanju ne dajejo v mtabelo in zato jih tudi tukaj ne
  // smemo dat noter!!!!
  // Ce bomo zeleli explicitno invocirati vgrajeno metodo, bomo napravili Wrapper




  ///////////////////////////////////////////////////////////////////////////////////////////



  ////////////////////////////////////////////////////
  // Object class
  ////////////////////////////////////////////////////    

  mString objName("Object");

  int startsWith = 0;

  // create classdata for Object: 3 methods, 1 class (no superclasses), no value (0), 
  // Object class address = 0 bacause we're creating Object itself, start
  int* classData = createPrimitiveClass( objName, 2, 1, 0, 0, startsWith );

  int* objectClass = classData;


  // add method 'clone': total 3 methods, startsWith, index in vtable (into method table) 0,
  // 0 - 1st method, code = 0
  addPrimitiveMethod(classData, 2, startsWith, 0, 0, 0 );  


  // add method 'print': total 3 methods, startsWith, index in vtable (into method table) 1,
  // 1 - 2nd method, code = 0
  addPrimitiveMethod(classData, 2, startsWith, 1, 1, object_print );


  // add method 'getClass': total 3 methods, startsWith, index in vtable (into method table) 1,
  // 1 - 2nd method, code = 0
  //addPrimitiveMethod(classData, 3, startsWith, 1, 1, object_getclass );


  primitiveClasses[0] = (int)classData;
  mLoadedClass* loaded = new mLoadedClass( &objName, (unsigned)classData, 2 );
  loadedClasses->append( loaded );  

 
  
  ////////////////////////////////////////////////////
  // Integer class
  ////////////////////////////////////////////////////

  objName = "Integer";

  startsWith = 0;

  // create classdata for Integer: 4 methods, 2 classes (Object superclass), int value (4), 
  classData = createPrimitiveClass( objName, 4, 2, 4, (int)objectClass, startsWith );


  // methods for Object

  // add method 'Object.clone': total 4 methods, startsWith, index in vtable (into method table) 0,
  // 0 - 1st method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 0, 0, 0 );  // Object.clone: index = 0  


  // add method 'Object.print': total 4 methods, startsWith, index in vtable (into method table) 1,
  // 1 - 2nd method, code = object_print  
  addPrimitiveMethod(classData, 4, startsWith, 3, 1, object_print ); // Object.print:index=3:Integer.print

  // methods for Integer

  // add method 'Integer.clone': total 4 methods, startsWith, index in vtable (into method table) 2,
  // 2 - 3rd method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 2, 2, 0 );  // Integer.clone: index = 2

  // add method 'Integer.print': total 4 methods, startsWith, index in vtable (into method table) 3,
  // 3 - 4th method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 3, 3, integer_print );  // Integer.print: index = 3


  
  primitiveClasses[1] = (int)classData;
  loaded = new mLoadedClass( &objName, (unsigned)classData, 2 );
  loadedClasses->append( loaded );
  

  
  ////////////////////////////////////////////////////
  // Double class
  ////////////////////////////////////////////////////
  
  objName = "Double";

  startsWith = 0;

  // create classdata for Double: 4 methods, 2 classes (Object superclasses), double value (8), 
  classData = createPrimitiveClass( objName, 4, 2, 8, (int)objectClass, startsWith );

  // methods for Object

  // add method 'Object.clone': total 4 methods, startsWith, index in vtable (into method table) 0,
  // 0 - 1st method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 0, 0, 0 );  // Object.clone: index = 0  


  // add method 'Object.print': total 4 methods, startsWith, index in vtable (into method table) 1,
  // 1 - 2nd method, code = object_print  
  addPrimitiveMethod(classData, 4, startsWith, 3, 1, object_print ); // Object.print:index=3:Double.print

  // methods for Double

  // add method 'Double.clone': total 4 methods, startsWith, index in vtable (into method table) 2,
  // 2 - 3rd method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 2, 2, 0 );  // Double.clone: index = 2

  // add method 'Double.print': total 4 methods, startsWith, index in vtable (into method table) 3,
  // 3 - 4th method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 3, 3, double_print );  // Double.print: index = 3


  
  primitiveClasses[2] = (int)classData;
  loaded = new mLoadedClass( &objName, (unsigned)classData, 2 );
  loadedClasses->append( loaded );



  
  


  ////////////////////////////////////////////////////
  // String class
  ////////////////////////////////////////////////////
  
  objName = "String";

  startsWith = 0;

  // create classdata for String: 4 methods, 2 classes (Object superclasses), int len + ptr to chars (8), 
  classData = createPrimitiveClass( objName, 4, 2, 8, (int)objectClass, startsWith );


  // methods for Object

  // add method 'Object.clone': total 4 methods, startsWith, index in vtable (into method table) 0,
  // 0 - 1st method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 0, 0, 0 );  // Object.clone: index = 0  


  // add method 'Object.print': total 4 methods, startsWith, index in vtable (into method table) 3,
  // 1 - 2nd method, code = object_print  
  addPrimitiveMethod(classData, 4, startsWith, 3, 1, object_print ); // Object.print:index=3:String.print

  // methods for String

  // add method 'String.clone': total 4 methods, startsWith, index in vtable (into method table) 2,
  // 2 - 3rd method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 2, 2, 0 );  // String.clone: index = 2

  // add method 'String.print': total 4 methods, startsWith, index in vtable (into method table) 3,
  // 3 - 4th method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 3, 3, string_print );  // String.print: index = 3


  
  primitiveClasses[3] = (int)classData;
  loaded = new mLoadedClass( &objName, (unsigned)classData, 2 );
  loadedClasses->append( loaded );


  
  ////////////////////////////////////////////////////
  // Boolean class
  ////////////////////////////////////////////////////
  
  objName = "Boolean";

  startsWith = 0;

  // create classdata for Integer: 4 methods, 2 classes (Object superclasses), int value (4), 
  classData = createPrimitiveClass( objName, 4, 2, 4, (int)objectClass, startsWith );


  // methods for Object

  // add method 'Object.clone': total 4 methods, startsWith, index in vtable (into method table) 0,
  // 0 - 1st method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 0, 0, 0 );  // Object.clone: index = 0  


  // add method 'Object.print': total 4 methods, startsWith, index in vtable (into method table) 3,
  // 1 - 2nd method, code = object_print  
  addPrimitiveMethod(classData, 4, startsWith, 3, 1, object_print ); // Object.print:index=3:Boolean.print

  // methods for Boolean

  // add method 'Boolean.clone': total 4 methods, startsWith, index in vtable (into method table) 2,
  // 2 - 3rd method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 2, 2, 0 );  // Boolean.clone: index = 2

  // add method 'Boolean.print': total 4 methods, startsWith, index in vtable (into method table) 3,
  // 3 - 4th method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 3, 3, boolean_print );  // Boolean.print: index = 3


  
  primitiveClasses[4] = (int)classData;
  loaded = new mLoadedClass( &objName, (unsigned)classData, 2 );
  loadedClasses->append( loaded );



  


  ////////////////////////////////////////////////////
  // Block class
  ////////////////////////////////////////////////////
  
  objName = "Block";

  startsWith = 0;

  // create classdata for Block: 4 methods, 2 classes (Object superclasses), values (20), 
  classData = createPrimitiveClass( objName, 5, 2, 20, (int)objectClass, startsWith );


  // methods for Object

  // add method 'Object.clone': total 5 methods, startsWith, index in vtable (into method table) 0,
  // 0 - 1st method, code = 0
  addPrimitiveMethod(classData, 5, startsWith, 0, 0, 0 );  // Object.clone: index = 0  


  // add method 'Object.print': total 5 methods, startsWith, index in vtable (into method table) 1,
  // 1 - 2nd method, code = object_print  
  addPrimitiveMethod(classData, 5, startsWith, 1, 1, object_print ); // Object.print:index=1:Object.print

  // methods for Block

  // add method 'Block.clone': total 5 methods, startsWith, index in vtable (into method table) 2,
  // 2 - 3rd method, code = 0
  //addPrimitiveMethod(classData, 5, startsWith, 2, 2, 0 );  // Block.clone: index = 2

  // add method 'Block.print': total 5 methods, startsWith, index in vtable (into method table) 1,
  // 3 - 4th method, code = 0
  //addPrimitiveMethod(classData, 5, startsWith, 3, 1, object_print );  // Block.print: index = 1

  // add method 'Block.exec': total 5 methods, startsWith, index in vtable (into method table) 4,
  // 4 - 5th method, code = 0


  // exec is not added, because it is not exposed as a high level method
  //addPrimitiveMethod(classData, 5, startsWith, 2, 2, 0 );  // Block.exec: index = 4


  ////////////////////////////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////////////
  // additional methods from native world for Block



  // sploh potrebno????????????   zdaj je testno
  addPrimitiveMethod(classData, 5, startsWith, 2, 2, block_isTopLevel );  // Block.isTopLevel: index = 4
  

  
  primitiveClasses[5] = (int)classData;
  loaded = new mLoadedClass( &objName, (unsigned)classData, 2 );
  loadedClasses->append( loaded );



  
  ////////////////////////////////////////////////////
  // While class
  ////////////////////////////////////////////////////
  
  objName = "While";

  startsWith = 0;

  // create classdata for Block: 5 methods, 2 classes (Object superclasses), values (12), 
  classData = createPrimitiveClass( objName, 4, 2, 12, (int)objectClass, startsWith );


  // methods for Object

  // add method 'Object.clone': total 5 methods, startsWith, index in vtable (into method table) 0,
  // 0 - 1st method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 0, 0, 0 );  // Object.clone: index = 0  


  // add method 'Object.print': total 5 methods, startsWith, index in vtable (into method table) 1,
  // 1 - 2nd method, code = object_print  
  addPrimitiveMethod(classData, 4, startsWith, 1, 1, object_print ); // Object.print:index=1:Object.print

  // methods for While

  // add method 'While.clone': total 5 methods, startsWith, index in vtable (into method table) 2,
  // 2 - 3rd method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 2, 2, 0 );  // While.clone: index = 2

  // add method 'While.print': total 5 methods, startsWith, index in vtable (into method table) 1,
  // 3 - 4th method, code = 0
  //addPrimitiveMethod(classData, 4, startsWith, 3, 1, object_print );  // While.print: index = 1

  // add method 'While.exec': total 5 methods, startsWith, index in vtable (into method table) 4,
  // 4 - 5th method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 3, 3, 0 );  // While.exec: index = 4


  
  primitiveClasses[6] = (int)classData;
  loaded = new mLoadedClass( &objName, (unsigned)classData, 2 );
  loadedClasses->append( loaded );



  

  ////////////////////////////////////////////////////
  // DoWhile class
  ////////////////////////////////////////////////////
  
  objName = "DoWhile";

  startsWith = 0;

  // create classdata for Block: 5 methods, 2 classes (Object superclasses), values (12), 
  classData = createPrimitiveClass( objName, 4, 2, 12, (int)objectClass, startsWith );


  // methods for Object

  // add method 'Object.clone': total 5 methods, startsWith, index in vtable (into method table) 0,
  // 0 - 1st method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 0, 0, 0 );  // Object.clone: index = 0  


  // add method 'Object.print': total 5 methods, startsWith, index in vtable (into method table) 1,
  // 1 - 2nd method, code = object_print  
  addPrimitiveMethod(classData, 4, startsWith, 1, 1, object_print ); // Object.print:index=1:Object.print

  // methods for DoWhile

  // add method 'DoWhile.clone': total 5 methods, startsWith, index in vtable (into method table) 2,
  // 2 - 3rd method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 2, 2, 0 );  // DoWhile.clone: index = 2

  // add method 'DoWhile.print': total 5 methods, startsWith, index in vtable (into method table) 1,
  // 3 - 4th method, code = 0
  //addPrimitiveMethod(classData, 4, startsWith, 3, 1, object_print );  // DoWhile.print: index = 1

  // add method 'DoWhile.exec': total 5 methods, startsWith, index in vtable (into method table) 4,
  // 4 - 5th method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 3, 3, 0 );  // DoWhile.exec: index = 4


  
  primitiveClasses[7] = (int)classData;
  loaded = new mLoadedClass( &objName, (unsigned)classData, 2 );
  loadedClasses->append( loaded );


  
  ////////////////////////////////////////////////////
  // If class
  ////////////////////////////////////////////////////
  
  objName = "If";

  startsWith = 0;

  // create classdata for Block: 5 methods, 2 classes (Object superclasses), values (16), 
  classData = createPrimitiveClass( objName, 4, 2, 16, (int)objectClass, startsWith );


  // methods for Object

  // add method 'Object.clone': total 5 methods, startsWith, index in vtable (into method table) 0,
  // 0 - 1st method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 0, 0, 0 );  // Object.clone: index = 0  


  // add method 'Object.print': total 5 methods, startsWith, index in vtable (into method table) 1,
  // 1 - 2nd method, code = object_print  
  addPrimitiveMethod(classData, 4, startsWith, 1, 1, object_print ); // Object.print:index=1:Object.print

  // methods for If

  // add method 'If.clone': total 5 methods, startsWith, index in vtable (into method table) 2,
  // 2 - 3rd method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 2, 2, 0 );  // If.clone: index = 2

  // add method 'If.print': total 5 methods, startsWith, index in vtable (into method table) 1,
  // 3 - 4th method, code = 0
  //addPrimitiveMethod(classData, 4, startsWith, 3, 1, object_print );  // If.print: index = 1

  // add method 'If.exec': total 5 methods, startsWith, index in vtable (into method table) 4,
  // 4 - 5th method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 4, 4, 0 );  // If.exec: index = 4


  
  primitiveClasses[8] = (int)classData;
  loaded = new mLoadedClass( &objName, (unsigned)classData, 2 );
  loadedClasses->append( loaded );



  

  ////////////////////////////////////////////////////
  // For class
  ////////////////////////////////////////////////////
  
  objName = "For";

  startsWith = 0;

  // create classdata for Block: 5 methods, 2 classes (Object superclasses), values (20), 
  classData = createPrimitiveClass( objName, 4, 2, 20, (int)objectClass, startsWith );


  // methods for Object

  // add method 'Object.clone': total 5 methods, startsWith, index in vtable (into method table) 0,
  // 0 - 1st method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 0, 0, 0 );  // Object.clone: index = 0  


  // add method 'Object.print': total 5 methods, startsWith, index in vtable (into method table) 1,
  // 1 - 2nd method, code = object_print  
  addPrimitiveMethod(classData, 4, startsWith, 1, 1, object_print ); // Object.print:index=1:Object.print

  // methods for For

  // add method 'For.clone': total 5 methods, startsWith, index in vtable (into method table) 2,
  // 2 - 3rd method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 2, 2, 0 );  // For.clone: index = 2

  // add method 'For.print': total 5 methods, startsWith, index in vtable (into method table) 1,
  // 3 - 4th method, code = 0
  //addPrimitiveMethod(classData, 4, startsWith, 3, 1, object_print );  // For.print: index = 1

  // add method 'For.exec': total 5 methods, startsWith, index in vtable (into method table) 4,
  // 4 - 5th method, code = 0
  addPrimitiveMethod(classData, 4, startsWith, 4, 4, 0 );  // For.exec: index = 4


  
  primitiveClasses[9] = (int)classData;
  loaded = new mLoadedClass( &objName, (unsigned)classData, 2 );
  loadedClasses->append( loaded );  

}



int* mVirtualMachine::createPrimitiveClass(mString& objName, int nMethods, int nObjects, int valueSize,
					   int objectClassData, int& endsWith)
{  

  // first is info for objects of this class
  //info for Object: object_size, mtable_offset, initial selfish, ctable_size, meta_size

  int descriptorSize = 8;     // descriptor is 8 bytes (self & selfish)
  int mtableEntrySize = 12;   // method table entry is 12 bytes (address, env, and actual self)
  int ctableEntrySize = 12;   // class table entry is 12 bytes (address, selfish, and mtable offset)
  int vtableEntrySize = 4;    // entry in virtual table is 4 bytes (index into method table)

  int infoSize = 20;          // object info is 20 bytes
  int vtableSize = vtableEntrySize * nMethods;  // size of virtual table in bytes
  int mtableSize = mtableEntrySize * nMethods;  // size of method table in bytes
  int ctableSize = ctableEntrySize * nObjects;  // size of class table in bytes


  int nObjectMethods = 2;   // class Object has 2 methods


  // size of virtual table for Object class is # of Object methods * vtable entry size
  // this is needed in order to calculate selfish offset of derived object (Integer, Double,...)
  int objectVtableSize = nObjectMethods * vtableEntrySize;

  // size of combined vtable
  int totalVtableSize = objectVtableSize + nMethods * vtableEntrySize;

  // this is needed in order to calculate mtable offset of derived object (Integer, Double,...)
  int objectMtableSize = nObjectMethods * mtableEntrySize;

  // size of metadata in bytes
  // 5 entries (ptr to classname, # superclasses, superlist, # methods, metadata table for methods
  int metaSize = 5 * 4;


  // this is the size of runtime object ( classdata and meta ptrs, descriptors, method table, and
  // value inside object)
  // 8 bytes is for classdata ptr and metadata ptr
  int baseSize = 8 + descriptorSize * nObjects + mtableSize + valueSize;

  // size of classdata in bytes
  int classDataSize = infoSize + ctableSize + totalVtableSize + mtableSize;

  
  char *classSpace = new char[classDataSize  + metaSize];
  int *data = (int*)classSpace;
  
  int *classData = data + 5;  // classdata points to actual class data which begins after metadata
  

  /////////////////////////////////////////////////////////////////
  // first store meta data
  /////////////////////////////////////////////////////////////////

  // 1st meta field is ptr to classname
  int namelen = objName.length();
  char *className = new char[namelen + 1];   // +1 for '\0'
  memcpy(className, objName, namelen);
  className[namelen] = '\0';   // store null

  // store ptr to classname
  data[0] = (int)className;

  // store # of superclasses ( -1 because nObjects is # of classes in this class, which is 1 for Object,
  // and 2 for other classes )
  data[1] = nObjects - 1;

  // store ptr to array of superclass addresses

  int* superAddresses = new int[nObjects - 1];
  
  // the only superclass in primitive classes is Object
  if(objectClassData != 0)  // we're not creating Object class, so superclass must be Object
    *superAddresses = objectClassData;
    

  data[2] = (int)superAddresses;
  

  //TUKAJ dodaj meta data za metode...ali spodaj RAJE ...


  /////////////////////////////////////////////////////////////////
  // ordinary class data
  /////////////////////////////////////////////////////////////////

  classData[0] = baseSize;        // size of object at runtime
  classData[1] = infoSize + ctableSize + vtableSize; // offset of method table
  classData[2] = infoSize + ctableSize;  // initial selfish goes after info, after classtable
  classData[3] = nObjects;               // # entries in classtable
  classData[4] = metaSize;               // meta info size

 
  // object table format is | class address | selfish offset | mtable offset |)

  // class table entry for Object is always first
  if(objectClassData == 0)   // we're creating data for Object itself
    classData[5] =  (int)classData;
  else
    classData[5] = objectClassData;

  classData[6] = infoSize + ctableSize; // selfish offset for Object
  classData[7] = infoSize + ctableSize + descriptorSize * nObjects;  // mtable offset

  endsWith = 8;   // where virtual table starts

  // second comes class table entry for derived object (Integer, Double, ...)
  if(objectClassData != 0){  // we're creating data for a derived Object...

    classData[8] = (int)classData; classData[9] = infoSize + ctableSize + objectVtableSize;
    classData[10] = infoSize + ctableSize + totalVtableSize + objectMtableSize;    
    endsWith = 11;

  }

  return classData;

  // virtual table
  //data_object[start + 0] = 0;   // index of 'clone' 
  //data_object[start + 1] = 1;   // index of 'print'



  //mLoadedClass* loaded = new mLoadedClass( objName, (unsigned)classData, 2 );
  //loadedClasses->append( loaded );

}


void mVirtualMachine::addPrimitiveMethod(int* classData, int nTotalMethods, int start, int index,
					 int methodNumber, char* methodCode)
{

  // virtual table index
  classData[start+methodNumber] = index;


  int mstart = start + nTotalMethods;  // start of method table is after vtable 

  // method table

  classData[mstart+3*methodNumber]   = (int)methodCode;
  classData[mstart+3*methodNumber+1] = 0;              // actual self
  classData[mstart+3*methodNumber+2] = 0;              // env ptr

}



int mVirtualMachine::getObjectClass()
{
  return primitiveClasses[0];
}

int mVirtualMachine::getIntegerClass()
{
  return primitiveClasses[1];
}

int mVirtualMachine::getDoubleClass()
{
  return primitiveClasses[2];
}

int mVirtualMachine::getStringClass()
{
  return primitiveClasses[3];
}

int mVirtualMachine::getBooleanClass()
{
  return primitiveClasses[4];
}


int mVirtualMachine::getBlockClass()
{
  return primitiveClasses[5];
}

int mVirtualMachine::getWhileClass()
{
  return primitiveClasses[6];
}


int mVirtualMachine::getDoWhileClass()
{
  return primitiveClasses[7];
}


int mVirtualMachine::getIfClass()
{
  return primitiveClasses[8];
}

int mVirtualMachine::getForClass()
{
  return primitiveClasses[9];
}





int mVirtualMachine::appendInstruction(char *instruction,int pos,short opc, int dest, int l1, int l2, 
				       int l3,int l4)
{

  // calculate instruction size
  int size = 2;   // opcode which is 16 bits, is always present  

  // copy opcode
  memcpy( instruction+pos, &opc, 2 );


  if( dest != -1 ){
    memcpy( instruction +pos+ 2, &dest, 4 );  
    size += 4;
  }
  if( l1 != -1 ){
    memcpy( instruction +pos+ 6, &l1, 4 );
    size += 4;
  }
  if( l2 != -1 ){
    memcpy( instruction +pos+ 10, &l2, 4 );
    size += 4;
  }
  if( l3 != -1 ){
    memcpy( instruction +pos+ 14, &l3, 4 );
    size += 4;
  }
  if( l4 != -1 ){
    memcpy( instruction +pos+ 18, &l4, 4 );
    size += 4;
  }
  return size;
}



void mVirtualMachine::loadNativeLibs()
{

  // base library 

  mString libName = vmParameters->getLibDirectory() + "base.so";

  void *handle = dlopen( libName, RTLD_LAZY );
  


  if(!handle){
    cout<<"Error opening native library: "<<dlerror()<<endl;
    exit(1);
  }


  // add it to array
  nativeHandles[ nNativeHandles ] = handle;

  nNativeHandles++;
  

}


void mVirtualMachine::unloadNativeLibs()
{

  // base lib

  for(int i = 0; i < nNativeHandles; i++){
    
    dlclose( nativeHandles[i] );  

  }
  
}
