#ifndef __MVIRTUAL_MACHINE_H
#define __MVIRTUAL_MACHINE_H


#include "mString.h"
#include "mList.h"
#include "mLoadedClass.h"
#include "mTestInterpreter.h"
#include <iostream>
#include "mVMEnvironment.h"
#include "mVMParameters.h"




// maximum number of shared libraries that can be loaded
#define MAX_NATIVE_HANDLES  250



using namespace std;







class mVirtualMachine{
 public:
  mVirtualMachine(char**);
  virtual ~mVirtualMachine();

  // this is the primary class loading routine
  // returns loaded class or 0 on error
  mLoadedClass* loadLocalClass(mString*);

  // executes previously loaded class
  void executeClass(mLoadedClass*);

  // returns class with specified name or 0 if it's not loaded
  mLoadedClass* getLoadedClass(mString*);

  // returns class with specified address or 0 if it's not loaded
  mLoadedClass* getLoadedClass(int);

  
  // returns class with specified name. if the class is not loaded, it loads it
  mLoadedClass* findAndLoadClass(mString*);

  // returns native lib handle
  void *getNativeLibraryHandle();

  // allocates size bytes of memory
  void* allocate(int);

  // returns start of stack
  int* getStack(){ return stack; }

  // returns stack size
  int getStackSize(){ return stackSize; }

  // loads classes for primitive types such as Object, Integer, String, ...
  void loadPrimitiveClasses();

  // returns address of class Object
  int getObjectClass();

  // returns address of class Integer
  int getIntegerClass();

  // returns address of class Double
  int getDoubleClass();

  // returns address of class String
  int getStringClass();

  // returns address of class Boolean
  int getBooleanClass();

  // returns address of class Block
  int getBlockClass();

  // returns address of class While
  int getWhileClass();

  // returns address of class DoWhile
  int getDoWhileClass();

  // returns address of class If
  int getIfClass();

  // returns address of class For
  int getForClass();


  // returns the number of opened handles for native libraries
  int numNativeHandles(){ return nNativeHandles; }

  void** getNativeHandles() { return nativeHandles; }


  // creates object of primitive type from classdata
  // used only for builtin types !!!
  int* create_object(int*);

  // creates object of user type from classdata
  // used only for user types !!!
  // second argument is used to add it to object size
  // this is useful when we need additional size of object
  // than it's specified in class info (class Class for example gets additional
  // value of classaddress when it is instantiated)
  int* create_user_object(int*,int=0);


  // returns this virtual machine's environment
  mVMEnvironment* getEnvironment(){ return environment; }


  
  // returns interpreter for this virtual machine
  mAbstractInterpreter* getInterpreter(){ return interpreter; }

 private:

  // 10k entries for stack
  static const int stackSize = 10000;


  // offset of vtable for Object for all primitive types except object
  // vtable starts after classinfo (20 bytes) and after object table, which
  // has 2 entries for all primitive types (each entry 12 bytes)
  static const int object_vtable_offset = 20 + 24;

  // virtual table for inherited primitive class begins after virtual table for Object
  // which is +8 because vtable of Object is 8 bytes (2 methods)
  static const int second_vtable_offset = 20 + 24 + 8;  // after vtable for object

  // reads meta data from classfile and sets it up as pointers
  // classdata buffer to store meta data pointers to, file ptr (which is already at the
  // beginning of meta data)
  // returns number of bytes read
  int readMetaInfo(int*,classHeader&, FILE*&);

  // relocates class to specified memory address using relocation table
  // args: class code, address to relocate to, relocation table, num relocation entries
  void relocateClass(char*,int,int*,int);

  // relocates class' method table
  void relocateMethodTable(char*, int* ,int, int);

  // links class with external class
  // args:  class code, external class address, link offsets, # of link offsets
  void linkExternalClass(char*, int, int*, int);

  int appendInstruction(char*,int,short,int=-1, int=-1,int=-1, int=-1,int=-1);


  // loads libraries for native functions
  void loadNativeLibs();

  // unloads...
  void unloadNativeLibs();

  // creates primitive class of specified name, # of methods, # of objects in this object (1 for Object, 
  // and 2 for derived), value size (4 for Integer, 8 for Double, 0 for Object,...), classdata for
  // object (0 when creating classdata for Object itself), &where virtual table starts after data
  // returns ptr to created classdata
  int* createPrimitiveClass( mString&, int, int, int, int, int& );

  // adds method address and index to classdata of a primitive object
  // args: classdata, # of total methods in classdata, start of virtual table from start of classdata,
  // index in vtable, number of method (starting from 0), method code
  void addPrimitiveMethod(int*, int, int, int, int, char*);


  // adds object table entry for a primitive class
  //void addPrimitiveObjectTable(char* classData, char* otherClassAddress, int );


  // sets environment for vm
  

  // all loaded classes
  mList<mLoadedClass>* loadedClasses;

  // virtual stack
  int* stack;

  // memory pool for object allocation
  char* originalHeap;
  char* vmHeap;

  // handles for native library
  void** nativeHandles;


  // # of native handles
  int nNativeHandles;

  // addresses of primitive classes
  int primitiveClasses[10];


  // environment of this virtual machine
  mVMEnvironment* environment;

  // the params
  mVMParameters* vmParameters;

  // the interpreter
  mAbstractInterpreter* interpreter;  

};

#endif
