%option yylineno

%{

#include <iostream>
#include "mString.h"
#include "mField.h"
#include "mParameter.h"
#include "mEnvironment.h"
#include "mStack.h"
#include "mParseTree.h"
#include "mLiteral.h"
#include "mCode.h"
#include "mClassExtension.h"


#include "y.tab.h"

using namespace std;



// for buffer switching
#define MAX_INCLUDE_DEPTH   10

struct buffer{
    mString *fileName;
    YY_BUFFER_STATE state;
    int yylineno;
};


buffer include_stack[MAX_INCLUDE_DEPTH];

int include_stack_ptr=0;

// current file
mString* currentBufferName=0;

extern mString thisClassFile;

bool switchBuffer(mString*);

bool deleteBuffer();

bool ignoreLF=true;     // ignore \n


// additional float formats
//Float2          [\.]{Digit}+{ExponentPart}?{FlSuffix}?
//Float3          {Digit}+{ExponentPart}{FlSuffix}?
//Float4          {Digit}+{FlSuffix}
//FloatingPoint   ({Float1}|{Float2}|{Float3}|{Float4})


%}





Separator       [\(\)\{\}\[\;\]\,\.]
Delimiter1      [\=\>\<\!\~\?\:\+\-\*\/\&\|\^\%]
HexDigit        [0-9a-fA-F]
Digit           [0-9]
OctalDigit      [0-7]
TetraDigit      [0-3]
NonZeroDigit    [1-9]
Letter          [a-zA-Z_]
AnyButSlash     [^\/]
AnyButAstr      [^\*]
BLANK           [ ]
BLK             [\b]
TAB             [\t]
FF              [\f]
ESCCHR          [\\]
CR              [\r]
LF              [\n]
UniEsc          [\1b]

OctEscape1      [\\]{OctalDigit}
OctEscape2      [\\]{OctalDigit}{OctalDigit}
OctEscape3      [\\]{TetraDigit}{OctalDigit}{OctalDigit}
OctEscape       ({OctEscape1}|{OctEscape2}|{OctEscape3})

Escape          [\\]([r]|[n]|[b]|[f]|[t]|[\\]|[\']|[\"])
ULetter         ({Letter}|{UniEsc})
Identifier      {ULetter}({ULetter}|{Digit})*

Comment1        [\/][\*]({AnyButAstr}|[\*]{AnyButSlash})*[\*][\/]
Comment2        [\/][\/].*
Comment         ({Comment1}|{Comment2})

Dimension       [\[]({CR}|{LF}|{FF}|{TAB}|{BLK}|{BLANK}|{Comment})*[\]]

IntSuffix       ([l]|[L])
DecimalNum      {NonZeroDigit}{Digit}*{IntSuffix}?
OctalNum        [0]{OctalDigit}*{IntSuffix}?
HexNum          [0]([x]|[X]){HexDigit}{HexDigit}*{IntSuffix}?
IntegerLiteral  ({DecimalNum}|{OctalNum}|{HexNum})

Sign            ([\+]|[\-])
FlSuffix        ([f]|[F]|[d]|[D])
SignedInt       {Sign}?{Digit}+
Expo            ([e]|[E])
ExponentPart    {Expo}{SignedInt}?
Float1          {Digit}+[\.]{Digit}+{ExponentPart}?{FlSuffix}?

FloatingPoint   {Float1}

AnyChrChr       [^\\']
AnyStrChr       [^\\"]
Character       [\']({Escape}|{OctEscape}|{AnyChrChr})[\']
String          [\"]({Escape}|{OctEscape}|{AnyStrChr})*[\"]
Numeric         ({IntegerLiteral}|{FloatingPoint})
Literal         ({Numeric}|{Character}|{String})

%%

{Separator}      { return yytext[0];  }
{Delimiter1}     { return yytext[0]; }
{String}         { 
                   mString str(yytext); 
                   // remove quotes from both sides
                   str = str.middle(1, str.length()- 2 );
                   yylval.stringVal = new mString( str );

                   return STRING_LITERAL; 
                 }

"category"       { return CATEGORY; }
"use"            { return USE; }
"class"          { return CLASS; }
"inherits"       { return INHERITS; }
"implements"     { return IMPLEMENTS; }
"constructor"    { return CONSTRUCTOR; }
"destructor"     { return DESTRUCTOR; }
"field"          { return FIELD; }
"method"         { return METHOD; }
"is"             { return IS; }
"this"           { return THIS; }
"new"            { return NEW; }
"super"          { return SUPER; }
"return"         { return RETURN; }
"while"          { return WHILE; }
"do"             { return DO; }
"as"             { return AS; }
"if"             { return IF; }
"then"           { return THEN; }
"else"           { return ELSE; }
"for"            { return FOR; }
"method"         { return METHOD; }
"public"         { return PUBLIC; }
"private"        { return PRIVATE; }
"protected"      { return PROTECTED; }
"restricted"     { return RESTRICTED; }
"native"         { return NATIVE; }
"static"         { return STATIC; }
"Void"           { return VOID; }
"Byte"           { return BYTE; }
"Character"      { return CHAR; }
"Short"          { return SHORT; }
"Integer"        { return INT; }
"Long"           { return LONG; }
"Float"          { return FLOAT; }
"Double"         { return DOUBLE; }
"String"         { return STRING; }
"Boolean"        { return BOOLEAN; }
"Method"         { 
                    char c; 
                    char *yycopy = strdup( yytext );
                    int i;
                    while ( (c = yyinput()) == ' ' && c != EOF );  // eat space

                    unput( c );  // put it back                    
		    strcpy( yytext,  yycopy );   // restore yytext
		    free(yycopy);

                  
                    if( c == '(' || c == ':' ){                     
                     return METHODTYPE; 
                    }else{

                     yylval.stringVal=new mString(yytext); 
                     return IDENTIFIER;
                    }                  
                 }
"Self"           { return SELF; }
"true"           { return TRUE; }
"false"          { return FALSE; }
"nil"            { return NIL; }
"=="             { return OP_EQ; }
"!="             { return OP_NOT_EQ; }
">="             { return OP_GREATER_OR_EQ; }
"<="             { return OP_LESS_OR_EQ; }
"&&"             { return OP_AND; }
"||"             { return OP_OR; }
"<-"             { return OP_METHODASSIGN; }
"+="             { return OP_ADDASSIGN; }
"-="             { return OP_DECASSIGN; }
"*="             { return OP_MULASSIGN; }
"/="             { return OP_DIVASSIGN; }
"++"             { return OP_INCREMENT; }
"--"             { return OP_DECREMENT; }


{Identifier} {  yylval.stringVal=new mString(yytext); return IDENTIFIER; }
{DecimalNum}    { yylval.intVal=strtol(yytext,0,10); return INTEGER_LITERAL; }
{OctalNum}      { yylval.intVal=strtol(yytext,0,8); return INTEGER_LITERAL; }
{HexNum}        { yylval.intVal=strtol(yytext,0,16); return INTEGER_LITERAL; }
{FloatingPoint} { yylval.doubleVal=strtod(yytext,0); return DOUBLE_LITERAL; }


{CR}            {}
{LF}            { if(!ignoreLF) return '\n'; }
{FF}            {}
{TAB}           {}
{BLK}           {}
{BLANK}         {}

{Comment}       {}

<<EOF>>         { if(!deleteBuffer()) return 0;   }


.     printf( "Unrecognized character: %s\n", yytext );

%%

// switches buffer
bool switchBuffer(mString *fname){
    if(include_stack_ptr>=MAX_INCLUDE_DEPTH){
        cerr<<"includes nested too deeply"<<endl;
        return false;
    }


    //cout<<"--> opening file "<<*fname<<endl;
    FILE* fp = fopen(*fname,"r");
    if(!fp) return false;
    
    //cout<<"--> file "<<*fname<<" opened!"<<endl;
    
        

    //cout<<"--> saving old buffername "<<*currentBufferName<<endl;

    include_stack[include_stack_ptr].fileName = currentBufferName;

    //cout<<"--> creating new buffername ";

    currentBufferName = new mString(*fname);

    //cout<<" "<<*currentBufferName<<endl;
    
    //cout<<"--> saving old yylineno "<<yylineno<<endl;
    include_stack[include_stack_ptr].yylineno = yylineno;


    //cout<<"--> saving old YY_CURRENT_BUFFER "<<YY_CURRENT_BUFFER<<endl;
    
    include_stack[include_stack_ptr++].state = YY_CURRENT_BUFFER;    

    //cout<<"--> creating new buffer!!!"<<endl;
    YY_BUFFER_STATE new_buffer = yy_create_buffer(fp, YY_BUF_SIZE);

    //cout<<"--> finally switching to new buffer!!!"<<endl;
    
    yy_switch_to_buffer( new_buffer );
    yyin = fp;
    return  true;
}



// deletes  buffer
bool deleteBuffer(){
     if(include_stack_ptr-1 < 0){
        //yyterminate();
        return false;
     }else{
        yy_delete_buffer(YY_CURRENT_BUFFER);	
		

	fclose(yyin);

        include_stack_ptr--;

	if(include_stack_ptr == 0) currentBufferName = &thisClassFile;
	else delete currentBufferName;
         
        currentBufferName = include_stack[include_stack_ptr].fileName;
        yylineno = include_stack[include_stack_ptr].yylineno;        

        yy_switch_to_buffer(include_stack[include_stack_ptr].state);
        return true;

     }
    return false;
}
